from __future__ import annotations

import datetime
import json as stdjson
import os
import re
import warnings
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from typing import Callable, Iterable, Any, Iterator, cast, \
    MutableMapping, Mapping, TypeVar, AsyncIterator

import aiofiles
from PIL import Image
from PIL.Image import DecompressionBombWarning
from aiofiles import os as aiofiles_os
from aiofiles.os import makedirs
from geopy import units
from geopy.format import DEGREES_FORMAT
from jinja2 import Environment as Jinja2Environment, select_autoescape, FileSystemLoader, pass_context, \
    pass_eval_context, Template as Jinja2Template, TemplateNotFound, BaseLoader
from jinja2.async_utils import auto_aiter, auto_await
from jinja2.filters import prepare_map, make_attrgetter
from jinja2.nodes import EvalContext
from jinja2.runtime import StrictUndefined, Context, Macro, DebugUndefined, new_context
from jinja2.utils import htmlsafe_json_dumps
from markupsafe import Markup, escape
from pdf2image.pdf2image import convert_from_path

from betty import _resizeimage, task
from betty.app import App
from betty.fs import hashfile, CACHE_DIRECTORY_PATH
from betty.functools import walk
from betty.html import CssProvider, JsProvider
from betty.locale import negotiate_localizeds, Localized, Datey, negotiate_locale, Date, DateRange, Localizer, \
    DEFAULT_LOCALIZER, Localey, get_data, Localizable
from betty.model import Entity, get_entity_type_name, GeneratedEntityId, UserFacingEntity, get_entity_type, \
    AncestryEntityId
from betty.model.ancestry import File, Citation, HasLinks, HasFiles, Subject, Witness, Dated, is_private, is_public, \
    AnonymousCitation, AnonymousSource
from betty.os import link_or_copy
from betty.path import rootname
from betty.project import ProjectConfiguration
from betty.render import Renderer
from betty.serde.dump import Dumpable, DictDump, VoidableDump, Void, minimize, none_void, void_none, Dump
from betty.string import camel_case_to_snake_case, camel_case_to_kebab_case, upper_camel_case_to_lower_camel_case

T = TypeVar('T')


def context_app(context: Context) -> App:
    return cast(Environment, context.environment).app


def context_task_context(context: Context) -> task.Context | None:
    task_context = context.resolve_or_missing('task_context')
    return task_context if isinstance(task_context, task.Context) else None


def context_localizer(context: Context) -> Localizer:
    localizer = context.resolve_or_missing('localizer')
    if isinstance(localizer, Localizer):
        return localizer
    raise RuntimeError('No `localizer` context variable exists in this Jinja2 template.')


class _Citer:
    def __init__(self):
        self._citations: list[Citation] = []
        self._anonymous_source = AnonymousSource()
        self._anonymous_citations: dict[AncestryEntityId | None, Citation] = {}

    def __iter__(self) -> enumerate[Citation]:
        return enumerate(self._citations, 1)

    def __len__(self) -> int:
        return len(self._citations)

    def cite(self, citation: Citation) -> int:
        if citation.private:
            source_key = None if citation.source is None else citation.source.ancestry_id
            try:
                citation = self._anonymous_citations[source_key]
            except KeyError:
                citation = AnonymousCitation(
                    source=citation.source or self._anonymous_source,
                )
                self._anonymous_citations[source_key] = citation
        if citation not in self._citations:
            self._citations.append(citation)
        return self._citations.index(citation) + 1


class _Breadcrumb(Dumpable):
    def __init__(self, label: str, url: str):
        self._label = label
        self._url = url

    def dump(self) -> DictDump[Dump]:
        return {
            '@type': 'ListItem',
            'name': self._label,
            'item': self._url,
        }


class _Breadcrumbs(Dumpable):
    def __init__(self):
        self._breadcrumbs: list[_Breadcrumb] = []

    def append(self, label: str, url: str) -> None:
        self._breadcrumbs.append(_Breadcrumb(label, url))

    def dump(self) -> VoidableDump:
        if not self._breadcrumbs:
            return Void
        return {
            '@context': 'https://schema.org',
            '@type': 'BreadcrumbList',
            'itemListElement': [
                {
                    'position': position,
                    **breadcrumb.dump(),
                }
                for position, breadcrumb
                in enumerate(self._breadcrumbs, 1)
            ],
        }


class EntityContexts:
    def __init__(self, *entities: Entity) -> None:
        self._contexts: dict[type[Entity], Entity | None] = defaultdict(lambda: None)
        for entity in entities:
            self._contexts[entity.type] = entity

    def __getitem__(self, entity_type_or_type_name: type[Entity] | str) -> Entity | None:
        if isinstance(entity_type_or_type_name, str):
            entity_type = get_entity_type(entity_type_or_type_name)
        else:
            entity_type = entity_type_or_type_name
        return self._contexts[entity_type]

    def __call__(self, *entities: Entity) -> EntityContexts:
        updated_contexts = EntityContexts()
        for entity in entities:
            updated_contexts._contexts[entity.type] = entity
        return updated_contexts


class Jinja2Provider:
    @property
    def globals(self) -> dict[str, Any]:
        return {}

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {}

    @property
    def tests(self) -> dict[str, Callable[..., bool]]:
        return {}


class Template(Jinja2Template):
    environment: Environment

    def new_context(
        self,
        vars: dict[str, Any] | None = None,
        shared: bool = False,
        locals: Mapping[str, Any] | None = None,
    ) -> Context:
        return new_context(
            self.environment,
            self.name,
            self.blocks,
            vars,
            shared,
            {
                'citer': _Citer(),
                'breadcrumbs': _Breadcrumbs(),
                **self.globals,
            },
            locals,
        )


class Environment(Jinja2Environment):
    template_class = Template
    globals: dict[str, Any]
    filters: dict[str, Callable[..., Any]]
    tests: dict[str, Callable[..., bool]]

    def __init__(self, app: App):
        template_directory_paths = [str(path / 'templates') for path, _ in app.assets.paths]
        super().__init__(
            loader=FileSystemLoader(template_directory_paths),
            enable_async=True,
            undefined=DebugUndefined if app.project.configuration.debug else StrictUndefined,
            autoescape=select_autoescape(['html.j2']),
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=[
                'jinja2.ext.do',
                'jinja2.ext.i18n',
            ],
        )

        self.app = app

        if app.project.configuration.debug:
            self.add_extension('jinja2.ext.debug')

        self._init_i18n()
        self._init_globals()
        self._init_filters()
        self._init_tests()
        self._init_extensions()

    def _init_i18n(self) -> None:
        self.install_gettext_callables(  # type: ignore[attr-defined]
            gettext=self._gettext,
            ngettext=self._ngettext,
            pgettext=self._pgettext,
            npgettext=self._npgettext,
        )
        self.policies['ext.i18n.trimmed'] = True

    @pass_context
    def _gettext(self, context: Context, message: str) -> str:
        return context_localizer(context).gettext(message)

    @pass_context
    def _ngettext(self, context: Context, message_singular: str, message_plural: str, n: int) -> str:
        return context_localizer(context).ngettext(message_singular, message_plural, n)

    @pass_context
    def _pgettext(self, context: Context, gettext_context: str, message: str) -> str:
        return context_localizer(context).pgettext(gettext_context, message)

    @pass_context
    def _npgettext(self, context: Context, gettext_context: str, message_singular: str, message_plural: str, n: int) -> str:
        return context_localizer(context).npgettext(gettext_context, message_singular, message_plural, n)

    def _init_globals(self) -> None:
        self.globals['app'] = self.app
        today = datetime.date.today()
        self.globals['today'] = Date(today.year, today.month, today.day)
        # Ideally we would use the Dispatcher for this. However, it is asynchronous only.
        self.globals['public_css_paths'] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, CssProvider)
            for path in extension.public_css_paths
        ]
        self.globals['public_js_paths'] = [
            path
            for extension in self.app.extensions.flatten()
            if isinstance(extension, JsProvider)
            for path in extension.public_js_paths
        ]
        self.globals['path'] = os.path
        self.globals['entity_contexts'] = EntityContexts()
        self.globals['localizer'] = DEFAULT_LOCALIZER

    def _init_filters(self) -> None:
        self.filters['unique'] = _filter_unique
        self.filters['map'] = _filter_map
        self.filters['flatten'] = _filter_flatten
        self.filters['walk'] = _filter_walk
        self.filters['localize'] = _filter_localize
        self.filters['locale_get_data'] = get_data
        self.filters['negotiate_localizeds'] = _filter_negotiate_localizeds
        self.filters['sort_localizeds'] = _filter_sort_localizeds
        self.filters['select_localizeds'] = _filter_select_localizeds
        self.filters['negotiate_dateds'] = _filter_negotiate_dateds
        self.filters['select_dateds'] = _filter_select_dateds
        self.filters['json'] = _filter_json
        self.filters['tojson'] = _filter_tojson
        self.filters['paragraphs'] = _filter_paragraphs
        self.filters['format_datey'] = _filter_format_datey
        self.filters['format_degrees'] = _filter_format_degrees
        self.filters['url'] = _filter_url
        self.filters['static_url'] = self.app.static_url_generator.generate
        self.filters['file'] = _filter_file
        self.filters['image'] = _filter_image
        self.filters['entity_type_name'] = get_entity_type_name
        self.filters['camel_case_to_snake_case'] = camel_case_to_snake_case
        self.filters['camel_case_to_kebab_case'] = camel_case_to_kebab_case
        self.filters['upper_camel_case_to_lower_camel_case'] = upper_camel_case_to_lower_camel_case
        self.filters['void_none'] = void_none
        self.filters['none_void'] = none_void
        self.filters['minimize'] = minimize

    def _init_tests(self) -> None:
        self.tests['entity'] = lambda x: isinstance(x, Entity)
        self.tests['public'] = is_public
        self.tests['private'] = is_private
        self.tests['user_facing_entity'] = lambda x: isinstance(x, UserFacingEntity)

        def _build_test_entity_type(resource_type: type[Entity]) -> Callable[[Any], bool]:
            def _test_entity_type(x: Any) -> bool:
                return isinstance(x, resource_type)
            return _test_entity_type
        for entity_type in self.app.entity_types:
            self.tests[f'{camel_case_to_snake_case(get_entity_type_name(entity_type))}_entity'] = _build_test_entity_type(entity_type)
        self.tests['has_generated_entity_id'] = lambda x: isinstance(x, Entity) and isinstance(x.id, GeneratedEntityId) or isinstance(x, GeneratedEntityId)
        self.tests['has_links'] = lambda x: isinstance(x, HasLinks)
        self.tests['has_files'] = lambda x: isinstance(x, HasFiles)
        self.tests['starts_with'] = str.startswith
        self.tests['subject_role'] = lambda x: isinstance(x, Subject)
        self.tests['witness_role'] = lambda x: isinstance(x, Witness)
        self.tests['date_range'] = lambda x: isinstance(x, DateRange)

    def _init_extensions(self) -> None:
        for extension in self.app.extensions.flatten():
            if isinstance(extension, Jinja2Provider):
                self.globals.update(extension.globals)
                self.filters.update(extension.filters)
                self.tests.update(extension.tests)

    def negotiate_template(
        self,
        names: list[str],
        parent: str | None = None,
        globals: MutableMapping[str, Any] | None = None,
    ) -> Template:
        for name in names:
            with suppress(TemplateNotFound):
                return cast(Template, self.get_template(name, parent, globals))
        raise TemplateNotFound(names[-1], f'Cannot find any of the following templates: {", ".join(names)}.')


Template.environment_class = Environment


class Jinja2Renderer(Renderer):
    def __init__(self, environment: Environment, configuration: ProjectConfiguration):
        self._environment = environment
        self._configuration = configuration
        self._loaders: dict[Path, BaseLoader] = {}

    def get_loader(self, root_path: Path) -> BaseLoader:
        if root_path not in self._loaders:
            self._loaders[root_path] = FileSystemLoader(root_path)
        return self._loaders[root_path]

    @property
    def file_extensions(self) -> set[str]:
        return {'.j2'}

    async def render_file(
        self,
        file_path: Path,
        *,
        task_context: task.Context | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        file_destination_path = file_path.parent / file_path.stem
        data: dict[str, Any] = {}
        if task_context is not None:
            data['task_context'] = task_context
        if localizer is not None:
            data['localizer'] = localizer
        try:
            relative_file_destination_path = file_destination_path.relative_to(self._configuration.www_directory_path)
        except ValueError:
            pass
        else:
            resource = '/'.join(relative_file_destination_path.parts)
            if self._configuration.locales.multilingual:
                resource_parts = resource.lstrip('/').split('/')
                if resource_parts[0] in map(lambda x: x.alias, self._configuration.locales.values()):
                    resource = '/'.join(resource_parts[1:])
            data['page_resource'] = resource
        root_path = rootname(file_path)
        rendered = await self.get_loader(root_path).load(
            self._environment,
            '/'.join(Path(file_path).relative_to(root_path).parts),
            self._environment.globals,
        ).render_async(data)
        async with aiofiles.open(file_destination_path, 'w', encoding='utf-8') as f:
            await f.write(rendered)
        await aiofiles_os.remove(file_path)
        return file_destination_path


@pass_context
def _filter_url(
    context: Context,
    resource: Any,
    media_type: str | None = None,
    *args: Any,
    locale: Localey | None = None,
    **kwargs: Any,
) -> str:
    return context_app(context).url_generator.generate(
        resource,
        media_type or 'text/html',
        *args,
        locale=locale or context_localizer(context).locale,  # type: ignore[misc]
        **kwargs,
    )


@pass_context
def _filter_localize(
    context: Context,
    localizable: Localizable,
) -> str:
    return localizable.localize(context_localizer(context))


@pass_context
def _filter_format_datey(
    context: Context,
    datey: Datey,
) -> str:
    return context_localizer(context).format_datey(datey)


@pass_context
def _filter_json(context: Context, data: Any, indent: int | None = None) -> str:
    """
    Converts a value to a JSON string.
    """
    return stdjson.dumps(data, indent=indent, cls=(context_app(context).json_encoder))


@pass_context
def _filter_tojson(context: Context, data: Any, indent: int | None = None) -> str:
    """
    Converts a value to a JSON string safe for use in an HTML document.

    This mimics Jinja2's built-in JSON filter, but uses Betty's own JSON encoder.
    """
    return htmlsafe_json_dumps(data, indent=indent, dumps=lambda *args, **kwargs: _filter_json(context, *args, **kwargs))


async def _filter_flatten(values_of_values: Iterable[Iterable[T]]) -> AsyncIterator[T]:
    async for values in auto_aiter(values_of_values):
        async for value in auto_aiter(values):
            yield value


def _filter_walk(value: Any, attribute_name: str) -> Iterable[Any]:
    return walk(value, attribute_name)


_paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')


@pass_eval_context
def _filter_paragraphs(eval_ctx: EvalContext, text: str) -> str | Markup:
    """Converts newlines to <p> and <br> tags.

    Taken from http://jinja.pocoo.org/docs/2.10/api/#custom-filters."""
    result = u'\n\n'.join(u'<p>%s</p>' % p.replace('\n', Markup('<br>\n'))
                          for p in _paragraph_re.split(escape(text)))
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def _filter_format_degrees(degrees: int) -> str:
    arcminutes = units.arcminutes(degrees=degrees - int(degrees))
    arcseconds = units.arcseconds(arcminutes=arcminutes - int(arcminutes))
    format_dict = dict(
        deg='°',
        arcmin="'",
        arcsec='"',
        degrees=degrees,
        minutes=round(abs(arcminutes)),
        seconds=round(abs(arcseconds))
    )
    return DEGREES_FORMAT % format_dict  # type: ignore[no-any-return]


async def _filter_unique(value: Iterable[T]) -> AsyncIterator[T]:
    seen = []
    async for value in auto_aiter(value):
        if value not in seen:
            yield value
            seen.append(value)


@pass_context
async def _filter_map(context: Context, values: Iterable[Any], *args: Any, **kwargs: Any) -> Any:
    """
    Maps an iterable's values.

    This mimics Jinja2's built-in map filter, but allows macros as callbacks.
    """
    if len(args) > 0 and isinstance(args[0], Macro):
        func: Macro | Callable[[Any], bool] = args[0]
    else:
        func = prepare_map(context, args, kwargs)
    async for value in auto_aiter(values):
        yield await auto_await(func(value))


@pass_context
async def _filter_file(context: Context, file: File) -> str:
    app = context_app(context)
    task_context = context_task_context(context)
    task_id = f'filter_file:{file.id}'
    if task_context is None or task_context.claim(task_id):
        file_destination_path = app.project.configuration.www_directory_path / 'file' / file.id / 'file' / file.path.name
        await _do_filter_file(file.path, file_destination_path)

    return f'/file/{file.id}/file/{file.path.name}'


async def _do_filter_file(file_source_path: Path, file_destination_path: Path) -> None:
    await makedirs(file_destination_path.parent, exist_ok=True)
    await link_or_copy(file_source_path, file_destination_path)


@pass_context
async def _filter_image(
    context: Context,
    file: File,
    width: int | None = None,
    height: int | None = None,
) -> str:
    app = context_app(context)
    task_context = context_task_context(context)

    destination_name = '%s-' % file.id
    if height and width:
        destination_name += '%dx%d' % (width, height)
    elif height:
        destination_name += '-x%d' % height
    elif width:
        destination_name += '%dx-' % width
    else:
        raise ValueError('At least the width or height must be given.')

    file_directory_path = app.project.configuration.www_directory_path / 'file'

    if file.media_type:
        if file.media_type.type == 'image':
            task_callable = _execute_filter_image_image
            destination_name += file.path.suffix
        elif file.media_type.type == 'application' and file.media_type.subtype == 'pdf':
            task_callable = _execute_filter_image_application_pdf
            destination_name += '.' + 'jpg'
        else:
            raise ValueError('Cannot convert a file of media type "%s" to an image.' % file.media_type)
    else:
        raise ValueError('Cannot convert a file without a media type to an image.')

    task_id = f'filter_image:{file.id}:{width or ""}:{height or ""}'
    if task_context is None or task_context.claim(task_id):
        cache_directory_path = CACHE_DIRECTORY_PATH / 'image'
        await task_callable(file.path, cache_directory_path, file_directory_path, destination_name, width, height)

    destination_public_path = '/file/%s' % destination_name

    return destination_public_path


async def _execute_filter_image_image(
    file_path: Path,
    cache_directory_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    width: int | None,
    height: int | None,
) -> None:
    with warnings.catch_warnings():
        # Ignore warnings about decompression bombs, because we know where the files come from.
        warnings.simplefilter('ignore', category=DecompressionBombWarning)
        image = Image.open(file_path)
    try:
        await _execute_filter_image(image, file_path, cache_directory_path, destination_directory_path, destination_name, width, height)
    finally:
        image.close()


async def _execute_filter_image_application_pdf(
    file_path: Path,
    cache_directory_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    width: int | None,
    height: int | None,
) -> None:
    with warnings.catch_warnings():
        # Ignore warnings about decompression bombs, because we know where the files come from.
        warnings.simplefilter('ignore', category=DecompressionBombWarning)
        image = convert_from_path(file_path, fmt='jpeg')[0]
    try:
        await _execute_filter_image(image, file_path, cache_directory_path, destination_directory_path, destination_name, width, height)
    finally:
        image.close()


async def _execute_filter_image(
    image: Image,
    file_path: Path,
    cache_directory_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    width: int | None,
    height: int | None,
) -> None:
    await makedirs(destination_directory_path, exist_ok=True)
    cache_file_path = cache_directory_path / ('%s-%s' % (hashfile(file_path), destination_name))
    destination_file_path = destination_directory_path / destination_name

    try:
        await link_or_copy(cache_file_path, destination_file_path)
    except FileNotFoundError:
        await makedirs(cache_directory_path, exist_ok=True)
        with image:
            if width is not None:
                width = min(width, image.width)
            if height is not None:
                height = min(height, image.height)

            if width is not None and height is not None:
                converted = _resizeimage.resize_cover(image, (width, height))
            elif width is not None:
                converted = _resizeimage.resize_width(image, width)
            elif height is not None:
                converted = _resizeimage.resize_height(image, height)
            else:
                raise ValueError('Width and height cannot both be None.')
            converted.save(cache_file_path)
        await makedirs(destination_directory_path, exist_ok=True)
        await link_or_copy(cache_file_path, destination_file_path)


@pass_context
def _filter_negotiate_localizeds(context: Context, localizeds: Iterable[Localized]) -> Localized | None:
    return negotiate_localizeds(context_localizer(context).locale, list(localizeds))


@pass_context
def _filter_sort_localizeds(context: Context, localizeds: Iterable[Localized], localized_attribute: str, sort_attribute: str) -> Iterable[Localized]:
    get_localized_attr = make_attrgetter(
        context.environment, localized_attribute)
    get_sort_attr = make_attrgetter(context.environment, sort_attribute)

    def _get_sort_key(x: Localized) -> Any:
        return get_sort_attr(negotiate_localizeds(context_localizer(context).locale, get_localized_attr(x)))

    return sorted(localizeds, key=_get_sort_key)


@pass_context
def _filter_select_localizeds(context: Context, localizeds: Iterable[Localized], include_unspecified: bool = False) -> Iterable[Localized]:
    for localized in localizeds:
        if include_unspecified and localized.locale in {None, 'mis', 'mul', 'und', 'zxx'}:
            yield localized
        if localized.locale is not None and negotiate_locale(context_localizer(context).locale, {localized.locale}) is not None:
            yield localized


@pass_context
def _filter_negotiate_dateds(context: Context, dateds: Iterable[Dated], date: Datey | None) -> Dated | None:
    with suppress(StopIteration):
        return next(_filter_select_dateds(context, dateds, date))
    return None


@pass_context
def _filter_select_dateds(context: Context, dateds: Iterable[Dated], date: Datey | None) -> Iterator[Dated]:
    if date is None:
        date = context.resolve_or_missing('today')
    return filter(
        lambda dated: dated.date is None or dated.date.comparable and dated.date in date,
        dateds,
    )



import logging
from pathlib import Path
from typing import Any

import dill
import docker
from aiofiles.tempfile import TemporaryDirectory, AiofilesContextManagerTempDir
from docker.errors import DockerException

from betty.app import App
from betty.extension.nginx.artifact import generate_dockerfile_file, generate_configuration_file
from betty.extension.nginx.docker import Container
from betty.serve import NoPublicUrlBecauseServerNotStartedError, AppServer


class DockerizedNginxServer(AppServer):
    def __init__(self, app: App) -> None:
        super().__init__(
            # Create a new app so we can modify it later.
            dill.loads(dill.dumps(app))
        )
        self._container: Container | None = None
        self._output_directory: AiofilesContextManagerTempDir[None, Any, Any] | None = None

    async def start(self) -> None:
        from betty.extension import Nginx

        await super().start()
        logging.getLogger().info('Starting a Dockerized nginx web server...')
        self._output_directory = TemporaryDirectory()
        output_directory_name = await self._output_directory.__aenter__()
        nginx_configuration_file_path = Path(output_directory_name) / 'nginx.conf'
        docker_directory_path = Path(output_directory_name) / 'docker'
        dockerfile_file_path = docker_directory_path / 'Dockerfile'

        self._app.project.configuration.debug = True
        # Work around https://github.com/bartfeenstra/betty/issues/1056.
        self._app.extensions[Nginx].configuration.https = False

        await generate_configuration_file(
            self._app,
            destination_file_path=nginx_configuration_file_path,
            https=False,
            www_directory_path='/var/www/betty',
        )
        await generate_dockerfile_file(
            self._app,
            destination_file_path=dockerfile_file_path,
        )
        self._container = Container(
            self._app.project.configuration.www_directory_path,
            docker_directory_path,
            nginx_configuration_file_path,
        )
        await self._container.start()

    async def stop(self) -> None:
        if self._container is not None:
            await self._container.stop()
        if self._output_directory is not None:
            await self._output_directory.__aexit__(None, None, None)

    @property
    def public_url(self) -> str:
        if self._container is not None:
            return 'http://%s' % self._container.ip
        raise NoPublicUrlBecauseServerNotStartedError()

    @classmethod
    def is_available(cls) -> bool:
        try:
            docker.from_env().info()
            return True
        except DockerException as e:
            logging.getLogger().warning(e)
            return False



from __future__ import annotations

from textwrap import indent
from typing import cast, Any, Self

from betty.error import UserFacingError
from betty.locale import Str, Localizable, Localizer


class Requirement(Localizable):
    def is_met(self) -> bool:
        raise NotImplementedError(repr(self))

    def assert_met(self) -> None:
        if not self.is_met():
            raise RequirementError(self)
        return None

    def summary(self) -> Str:
        raise NotImplementedError(repr(self))

    def details(self) -> Str | None:
        return None

    def localize(self, localizer: Localizer) -> str:
        string = self.summary().localize(localizer)
        details = self.details()
        if details is not None:
            string += f'\n{"-" * len(string)}'
            string += f'\n{details.localize(localizer)}'
        return string

    def reduce(self) -> Requirement | None:
        """
        Removes unnecessary components of this requirement.
        - Collections can flatten unnecessary hierarchies.
        - Empty decorators or collections can 'dissolve' themselves and return None.

        This function MUST NOT modify self.
        """
        return self


class RequirementError(RuntimeError, UserFacingError):
    def __init__(self, requirement: Requirement):
        super().__init__(requirement)
        self._requirement = requirement

    def requirement(self) -> Requirement:
        return self._requirement


class RequirementCollection(Requirement):
    def __init__(self, *requirements: Requirement | None):
        super().__init__()
        self._requirements: list[Requirement] = [requirement for requirement in requirements if requirement]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._requirements == other._requirements

    def __add__(self, other: Any) -> Self:
        if not isinstance(other, Requirement):
            raise NotImplementedError(repr(self))
        self._requirements = [*self._requirements, other]
        return self

    def localize(self, localizer: Localizer) -> str:
        localized = super().localize(localizer)
        for requirement in self._requirements:
            localized += f'\n-{indent(requirement.localize(localizer), "  ")[1:]}'
        return localized

    def reduce(self) -> Requirement | None:
        reduced_requirements = []
        for requirement in self._requirements:
            reduced_requirement = requirement.reduce()
            if reduced_requirement:
                if type(reduced_requirement) is type(self):
                    reduced_requirements.extend(cast(RequirementCollection, reduced_requirement)._requirements)
                else:
                    reduced_requirements.append(reduced_requirement)
        if len(reduced_requirements) == 1:
            return reduced_requirements[0]
        if reduced_requirements:
            return type(self)(*reduced_requirements)
        return None


class AnyRequirement(RequirementCollection):
    def __init__(self, *requirements: Requirement | None):
        super().__init__(*requirements)
        self._summary = Str._('One or more of these requirements must be met')

    def is_met(self) -> bool:
        for requirement in self._requirements:
            if requirement.is_met():
                return True
        return False

    def summary(self) -> Str:
        return self._summary


class AllRequirements(RequirementCollection):
    def __init__(self, *requirements: Requirement | None):
        super().__init__(*requirements)
        self._summary = Str._('All of these requirements must be met')

    def is_met(self) -> bool:
        for requirement in self._requirements:
            if not requirement.is_met():
                return False
        return True

    def summary(self) -> Str:
        return self._summary



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout, QButtonGroup, QRadioButton, QWidget, QHBoxLayout, QLineEdit, \
    QFileDialog, QPushButton
from reactives.instance.property import reactive_property

from betty.app import App
from betty.app.extension import ConfigurableExtension
from betty.config import Configuration
from betty.extension.nginx.artifact import generate_configuration_file, generate_dockerfile_file
from betty.generate import Generator, GenerationContext
from betty.gui import GuiBuilder
from betty.gui.error import catch_exceptions
from betty.locale import Str
from betty.serde.dump import Dump, VoidableDump, minimize, Void, VoidableDictDump
from betty.serde.load import Asserter, Fields, OptionalField, Assertions
from betty.serve import ServerProvider, Server


class NginxConfiguration(Configuration):
    def __init__(self, www_directory_path: str | None = None, https: bool | None = None):
        super().__init__()
        self._https = https
        self.www_directory_path = www_directory_path

    @property
    @reactive_property
    def https(self) -> bool | None:
        return self._https

    @https.setter
    def https(self, https: bool | None) -> None:
        self._https = https

    @property
    @reactive_property
    def www_directory_path(self) -> str | None:
        return self._www_directory_path

    @www_directory_path.setter
    def www_directory_path(self, www_directory_path: str | None) -> None:
        self._www_directory_path = www_directory_path

    def update(self, other: Self) -> None:
        self._https = other._https
        self._www_directory_path = other._www_directory_path

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'https',
                Assertions(asserter.assert_or(asserter.assert_bool(), asserter.assert_none())) | asserter.assert_setattr(configuration, 'https'),
            ),
            OptionalField(
                'www_directory_path',
                Assertions(asserter.assert_str()) | asserter.assert_setattr(configuration, 'www_directory_path'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        dump: VoidableDictDump[VoidableDump] = {
            'https': self.https,
            'www_directory_path': Void if self.www_directory_path is None else str(self.www_directory_path),
        }
        return minimize(dump, True)


class _Nginx(ConfigurableExtension[NginxConfiguration], Generator, ServerProvider, GuiBuilder):
    @classmethod
    def label(cls) -> Str:
        return Str.plain('Nginx')

    @classmethod
    def description(cls) -> Str:
        return Str._('Generate <a href="">nginx</a> configuration for your site, as well as a <code>Dockerfile</code> to build a <a href="https://www.docker.com/">Docker</a> container around it.')

    @classmethod
    def default_configuration(cls) -> NginxConfiguration:
        return NginxConfiguration()

    @property
    def servers(self) -> Sequence[Server]:
        from betty.extension.nginx.serve import DockerizedNginxServer

        if DockerizedNginxServer.is_available():
            return [DockerizedNginxServer(self._app)]
        return []

    async def generate(self, task_context: GenerationContext) -> None:
        await generate_configuration_file(self._app)
        await generate_dockerfile_file(self._app)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @property
    def https(self) -> bool:
        if self._configuration.https is None:
            return self._app.project.configuration.base_url.startswith('https')
        return self._configuration.https

    @property
    def www_directory_path(self) -> str:
        return self._configuration.www_directory_path or str(self._app.project.configuration.www_directory_path)

    def gui_build(self) -> QWidget:
        return _NginxGuiWidget(self._app, self._configuration)


class _NginxGuiWidget(QWidget):
    def __init__(self, app: App, configuration: NginxConfiguration, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._app = app
        self._configuration = configuration
        layout = QFormLayout()

        self.setLayout(layout)

        https_button_group = QButtonGroup()

        def _update_configuration_https_base_url(checked: bool) -> None:
            if checked:
                self._configuration.https = None
        self._nginx_https_base_url = QRadioButton("Use HTTPS and HTTP/2 if the site's URL starts with https://")
        self._nginx_https_base_url.setChecked(self._configuration.https is None)
        self._nginx_https_base_url.toggled.connect(_update_configuration_https_base_url)
        layout.addRow(self._nginx_https_base_url)
        https_button_group.addButton(self._nginx_https_base_url)

        def _update_configuration_https_https(checked: bool) -> None:
            if checked:
                self._configuration.https = True
        self._nginx_https_https = QRadioButton('Use HTTPS and HTTP/2')
        self._nginx_https_https.setChecked(self._configuration.https is True)
        self._nginx_https_https.toggled.connect(_update_configuration_https_https)
        layout.addRow(self._nginx_https_https)
        https_button_group.addButton(self._nginx_https_https)

        def _update_configuration_https_http(checked: bool) -> None:
            if checked:
                self._configuration.https = False
        self._nginx_https_http = QRadioButton('Use HTTP')
        self._nginx_https_http.setChecked(self._configuration.https is False)
        self._nginx_https_http.toggled.connect(_update_configuration_https_http)
        layout.addRow(self._nginx_https_http)
        https_button_group.addButton(self._nginx_https_http)

        def _update_configuration_www_directory_path(www_directory_path: str) -> None:
            self._configuration.www_directory_path = None if www_directory_path == '' or www_directory_path == str(self._app.project.configuration.www_directory_path) else www_directory_path
        self._nginx_www_directory_path = QLineEdit()
        self._nginx_www_directory_path.setText(str(self._configuration.www_directory_path) if self._configuration.www_directory_path is not None else str(self._app.project.configuration.www_directory_path))
        self._nginx_www_directory_path.textChanged.connect(_update_configuration_www_directory_path)
        www_directory_path_layout = QHBoxLayout()
        www_directory_path_layout.addWidget(self._nginx_www_directory_path)

        @catch_exceptions
        def find_www_directory_path() -> None:
            found_www_directory_path = QFileDialog.getExistingDirectory(self, 'Serve your site from...', directory=self._nginx_www_directory_path.text())
            if '' != found_www_directory_path:
                self._nginx_www_directory_path.setText(found_www_directory_path)
        self._nginx_www_directory_path_find = QPushButton('...')
        self._nginx_www_directory_path_find.released.connect(find_www_directory_path)
        www_directory_path_layout.addWidget(self._nginx_www_directory_path_find)
        layout.addRow('WWW directory', www_directory_path_layout)



import re

_CAMEL_CASE_PATTERN = re.compile(r'(?<!^)(?=[A-Z])')


def camel_case_to_snake_case(string: str) -> str:
    return _CAMEL_CASE_PATTERN.sub('_', string).lower()


def camel_case_to_kebab_case(string: str) -> str:
    return _CAMEL_CASE_PATTERN.sub('-', string).lower()


def upper_camel_case_to_lower_camel_case(string: str) -> str:
    return string[0].lower() + string[1:]



class CssProvider:
    @property
    def public_css_paths(self) -> list[str]:
        raise NotImplementedError(repr(self))


class JsProvider:
    @property
    def public_js_paths(self) -> list[str]:
        raise NotImplementedError(repr(self))



from __future__ import annotations

from typing import Any, Iterable, Sized, TypeVar, Callable, Iterator, Generic, cast

T = TypeVar('T')
U = TypeVar('U')


def walk(item: Any, attribute_name: str) -> Iterable[Any]:
    child = getattr(item, attribute_name)

    # If the child has the requested attribute, yield it,
    if hasattr(child, attribute_name):
        yield child
        yield from walk(child, attribute_name)

    # Otherwise loop over the children and yield their attributes.
    try:
        child = iter(child)
    except TypeError:
        return
    for child in child:
        yield child
        yield from walk(child, attribute_name)


def slice_to_range(indices: slice, iterable: Sized) -> Iterable[int]:
    length = len(iterable)

    if indices.start is None:
        start = 0
    else:
        # Ensure the stop index is within range.
        start = max(-length, min(length, indices.start))

    if indices.stop is None:
        stop = max(0, length)
    else:
        # Ensure the stop index is within range.
        stop = max(-length, min(length, indices.stop))

    if indices.step is None:
        step = 1
    else:
        step = indices.step

    return range(start, stop, step)


class _Result(Generic[T]):
    def __init__(self, value: T | None, _error: BaseException | None = None):
        assert not _error or value is None
        self._value = value
        self._error = _error

    @property
    def value(self) -> T:
        if self._error:
            raise self._error
        return cast(T, self._value)

    def map(self, f: Callable[[T], U]) -> _Result[U]:
        if self._error:
            return cast(_Result[U], self)
        try:
            return _Result(f(self.value))
        except BaseException as e:
            return _Result(None, e)


def filter_suppress(raising_filter: Callable[[T], Any], exception_type: type[BaseException], items: Iterable[T]) -> Iterator[T]:
    for item in items:
        try:
            raising_filter(item)
            yield item
        except exception_type:
            continue



import logging

from betty.app import App


def getLogger() -> logging.Logger:
    return logging.getLogger(__name__)


class Loader:
    async def load(self) -> None:
        raise NotImplementedError(repr(self))


class PostLoader:
    async def post_load(self) -> None:
        raise NotImplementedError(repr(self))


async def load(app: App) -> None:
    await app.dispatcher.dispatch(Loader)()
    await app.dispatcher.dispatch(PostLoader)()



from __future__ import annotations

from betty.app.extension import Extension, UserFacingExtension
from betty.deriver import Deriver
from betty.load import PostLoader, getLogger
from betty.locale import Str
from betty.model.event_type import DerivableEventType


class _Deriver(UserFacingExtension, PostLoader):
    async def post_load(self) -> None:
        logger = getLogger()
        logger.info(self._app.localizer._('Deriving...'))

        deriver = Deriver(
            self.app.project.ancestry,
            self.app.project.configuration.lifetime_threshold,
            {
                event_type
                for event_type
                in self.app.event_types
                if issubclass(event_type, DerivableEventType)
            },
            localizer=self._app.localizer,
        )
        await deriver.derive()

    @classmethod
    def comes_before(cls) -> set[type[Extension]]:
        from betty.extension import Privatizer

        return {Privatizer}

    @classmethod
    def label(cls) -> Str:
        return Str._('Deriver')

    @classmethod
    def description(cls) -> Str:
        return Str._('Create events such as births and deaths by deriving their details from existing information.')



"""Integrates Betty with Python's setuptools."""

from setuptools import setup

from betty import _ROOT_DIRECTORY_PATH as ROOT_DIRECTORY_PATH
from betty._package import get_data_paths, find_packages

with open(ROOT_DIRECTORY_PATH / 'betty' / 'assets' / 'VERSION', encoding='utf-8') as f:
    VERSION = f.read()

with open(ROOT_DIRECTORY_PATH / 'README.md', encoding='utf-8') as f:
    long_description = f.read()


extras_require_pyinstaller = [
    'pyinstaller ~= 6.1, >= 6.1.0',
]


extras_require_setuptools = [
    'setuptools ~= 68.2, >= 68.2.2',
    'twine ~= 4.0, >= 4.0.0',
    'wheel ~= 0.40, >= 0.40.0',
]


extras_require_development = [
    'aioresponses ~= 0.7, >= 0.7.6',
    'autopep8 ~= 2.0, >= 2.0.2',
    'basedmypy ~= 2.0, >= 2.2.1',
    'codecov ~= 2.1, >= 2.1.12',
    'coverage ~= 7.2, >= 7.2.4',
    'flake8 ~= 6.0, >= 6.0.0',
    'html5lib ~= 1.1',
    'lxml ~= 4.9, >= 4.9.1; sys.platform != "win32"',
    'pip-licenses ~= 4.3, >= 4.3.0',
    'pytest ~= 7.3, >= 7.3.1',
    'pytest-aioresponses ~= 0.2, >= 0.2.0 ',
    'pytest-asyncio ~= 0.21, >= 0.21.0 ',
    'pytest-cov ~= 4.0, >= 4.0.0',
    'pytest-mock ~= 3.10, >= 3.10.0',
    'pytest-qt ~= 4.2, >= 4.2.0',
    'pytest-repeat ~= 0.9, >= 0.9.1',
    'pytest-xvfb ~= 3.0, >= 3.0.0',
    'types-aiofiles ~= 23.2, >= 23.2.0.0',
    'types-click ~= 7.1, >= 7.1.8',
    'types-mock ~= 5.0, >= 5.0.0.6',
    'types-polib ~= 1.2, >= 1.2.0.0',
    'types-pyyaml ~= 6.0, >= 6.0.6',
    'types-requests ~= 2.29, >= 2.29.0.0',
    'types-setuptools ~= 68.2, >= 68.2.0.0',
    *extras_require_pyinstaller,
    *extras_require_setuptools,
]


SETUP = {
    'name': 'betty',
    'description': 'Betty helps you visualize and publish your family history by building interactive genealogy websites out of your Gramps and GECOM family trees',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'version': VERSION,
    'license': 'GPLv3',
    'author': 'Bart Feenstra & contributors',
    'author_email': 'bart@mynameisbart.com',
    'url': 'https://github.com/bartfeenstra/betty',
    'project_urls': {
        'Github': 'https://github.com/bartfeenstra/betty',
        'Twitter': 'https://twitter.com/BettyProject',
        'X': 'https://twitter.com/BettyProject',
    },
    'classifiers': [
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: JavaScript',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: JavaScript',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Sociology :: Genealogy',
        'Topic :: Software Development :: Code Generators',
        'Natural Language :: Dutch',
        'Natural Language :: English',
        'Natural Language :: French',
        'Natural Language :: Ukrainian',
        'Typing :: Typed ',
    ],
    'python_requires': '~= 3.11',
    'install_requires': [
        'aiofiles ~= 23.2, >= 23.2.1',
        'aiohttp ~= 3.9',
        'babel ~= 2.12, >= 2.12.0',
        'click ~= 8.1, >= 8.1.2',
        'dill ~= 0.3, >= 0.3.6',
        'docker ~= 7.0, >= 7.0.0',
        'geopy ~= 2.3, >= 2.3.0',
        'jinja2 ~= 3.1, >= 3.1.1',
        'jsonschema ~= 4.17, >= 4.17.0',
        'langcodes ~= 3.3, >= 3.3.0',
        'markupsafe ~= 2.1, >= 2.1.1',
        'ordered-set ~= 4.1.0',
        'pdf2image ~= 1.16, >= 1.16.0',
        'polib ~= 1.2, >= 1.2.0',
        'Pillow ~= 10.1, >= 10.1.0',
        'PyQt6 ~= 6.5, >= 6.5.0',
        'pyyaml ~= 6.0, >= 6.0.0',
        'reactives ~= 0.5, >= 0.5.1',
    ],
    'extras_require': {
        'development': extras_require_development,
        'pyinstaller': extras_require_pyinstaller,
        'setuptools': extras_require_setuptools,
    },
    'entry_points': {
        'console_scripts': [
            'betty=betty.cli:main',
        ],
        'betty.extensions': [
            'betty.extension.CottonCandy=betty.extension.CottonCandy',
            'betty.extension.Demo=betty.extension.Demo',
            'betty.extension.Deriver=betty.extension.Deriver',
            'betty.extension.Gramps=betty.extension.Gramps',
            'betty.extension.Maps=betty.extension.Maps',
            'betty.extension.Privatizer=betty.extension.Privatizer',
            'betty.extension.HttpApiDoc=betty.extension.HttpApiDoc',
            'betty.extension.Trees=betty.extension.Trees',
            'betty.extension.Wikipedia=betty.extension.Wikipedia',
        ],
    },
    'packages': find_packages(),
    'package_data': {
        'betty': list(map(str, data_file_paths))
        for package, data_file_paths
        in get_data_paths().items()
    },
}

if __name__ == '__main__':
    setup(**SETUP)



from __future__ import annotations

from datetime import datetime
from typing import Iterator, TypeAlias

from betty.functools import walk
from betty.locale import DateRange, Date
from betty.model.ancestry import Person, Event, Citation, HasFiles, HasCitations, HasNotes, Source, \
    Presence, Privacy, HasMutablePrivacy, HasPrivacy

Expirable: TypeAlias = Person | Event | Date | None


class Privatizer:
    def __init__(self, lifetime_threshold: int):
        self._lifetime_threshold = lifetime_threshold
        self._seen: list[HasPrivacy] = []

    def privatize(self, subject: HasPrivacy) -> None:
        self._seen.clear()
        self._privatize(subject)

    def _privatize(self, subject: HasPrivacy) -> None:
        if subject.privacy is Privacy.PUBLIC:
            return

        if subject in self._seen:
            return
        self._seen.append(subject)

        if isinstance(subject, Person):
            self._privatize_person(subject)

        if isinstance(subject, Presence):
            self._privatize_presence(subject)

        if isinstance(subject, Event):
            self._privatize_event(subject)

        if isinstance(subject, Citation):
            self._privatize_citation(subject)

        if isinstance(subject, Source):
            self._privatize_source(subject)

        if isinstance(subject, HasCitations):
            self._privatize_has_citations(subject)

        if isinstance(subject, HasFiles):
            self._privatize_has_files(subject)

        if isinstance(subject, HasNotes):
            self._privatize_has_notes(subject)

    def _privatize_person(self, person: Person) -> None:
        # Do not change existing explicit privacy declarations.
        if person.privacy is Privacy.UNDETERMINED:
            person.privacy = self._person_privacy(person)

        if not person.private:
            return

        for name in person.names:
            self._mark_private(name)
            self._privatize(name)

        for presence in person.presences:
            self._privatize(presence)

    def _privatize_presence(self, presence: Presence) -> None:
        if not presence.private:
            return
        if presence.event is not None:
            self._mark_private(presence.event)
            self._privatize(presence.event)

    def _privatize_event(self, event: Event) -> None:
        if not event.private:
            return
        for presence in event.presences:
            self._privatize(presence)

    def _privatize_has_citations(self, has_citations: HasCitations & HasPrivacy) -> None:
        if not has_citations.private:
            return
        for citation in has_citations.citations:
            self._mark_private(citation)
            self._privatize(citation)

    def _privatize_citation(self, citation: Citation) -> None:
        if not citation.private:
            return
        if citation.source is not None:
            self._mark_private(citation.source)
            self._privatize(citation.source)

    def _privatize_source(self, source: Source) -> None:
        if not source.private:
            return
        for contained_source in source.contains:
            self._mark_private(contained_source)
            self._privatize(contained_source)
        for citation in source.citations:
            self._mark_private(citation)
            self._privatize(citation)

    def _privatize_has_files(self, has_files: HasFiles & HasPrivacy) -> None:
        if not has_files.private:
            return
        for file in has_files.files:
            self._mark_private(file)
            self._privatize(file)

    def _privatize_has_notes(self, has_notes: HasNotes & HasPrivacy) -> None:
        if not has_notes.private:
            return
        for note in has_notes.notes:
            self._mark_private(note)
            self._privatize(note)

    def _ancestors_by_generation(self, person: Person, generations_ago: int = 1) -> Iterator[tuple[Person, int]]:
        for parent in person.parents:
            yield parent, generations_ago
            yield from self._ancestors_by_generation(parent, generations_ago + 1)

    def _person_privacy(self, person: Person) -> Privacy:
        # A dead person is not private, regardless of when they died.
        if person.end is not None and person.end.event is not None:
            if person.end.event.date is None:
                return Privacy.PUBLIC
            if self.has_expired(person.end.event, 0):
                return Privacy.PUBLIC

        if self.has_expired(person, 1):
            return Privacy.PUBLIC

        for ancestor, generations_ago in self._ancestors_by_generation(person):
            if self.has_expired(ancestor, generations_ago + 1):
                return Privacy.PUBLIC

        # If any descendant has any expired event, the person is considered not private.
        for descendant in walk(person, 'children'):
            if self.has_expired(descendant, 1):
                return Privacy.PUBLIC

        return Privacy.PRIVATE

    def has_expired(
        self,
        subject: Expirable,
        generations_ago: int = 0,
    ) -> bool:
        if isinstance(subject, Person):
            return self._person_has_expired(subject, generations_ago)

        if isinstance(subject, Event):
            return self._event_has_expired(subject, generations_ago)

        if isinstance(subject, Date):
            return self._date_has_expired(subject, generations_ago)

        return False

    def _person_has_expired(self, person: Person, generations_ago: int) -> bool:
        for presence in person.presences:
            if presence.event is not None and self._event_has_expired(presence.event, generations_ago):
                return True
        return False

    def _event_has_expired(self, event: Event, generations_ago: int) -> bool:
        date = event.date

        if isinstance(date, DateRange):
            # We can only determine event expiration with certainty if we have an end date to work with. Someone born in
            # 2000 can have a valid birth event with a start date of 1800, which does nothing to help us determine
            # expiration.
            date = date.end

        return self.has_expired(date, generations_ago)

    def _date_has_expired(
        self,
        date: Date,
        generations_ago: int,
    ) -> bool:
        if not date.comparable:
            return False

        return date <= Date(datetime.now().year - self._lifetime_threshold * generations_ago, datetime.now().month, datetime.now().day)

    def _mark_private(self, target: HasMutablePrivacy) -> None:
        # Do not change existing explicit privacy declarations.
        if target.privacy is not Privacy.PUBLIC:
            target.privacy = Privacy.PRIVATE



from typing import Any

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QLabel


class Text(QLabel):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.setTextFormat(Qt.TextFormat.RichText)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard | Qt.TextInteractionFlag.LinksAccessibleByMouse | Qt.TextInteractionFlag.TextSelectableByKeyboard | Qt.TextInteractionFlag.TextSelectableByMouse)
        self.setOpenExternalLinks(True)


class Caption(Text):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        font = QFont()
        font.setPixelSize(12)
        self.setFont(font)



from __future__ import annotations

import platform
import sys
from importlib.metadata import distributions
from pathlib import Path
from typing import Iterator


def version() -> str | None:
    with open(Path(__file__).parent / 'assets' / 'VERSION', encoding='utf-8') as f:
        release_version = f.read().strip()
    if release_version == '0.0.0':
        return None
    return release_version


def version_label() -> str:
    return version() or 'development'


def is_stable() -> bool:
    return version() is not None


def is_development() -> bool:
    return version() is None


def _indent_mapping(items: dict[str, str]) -> str:
    max_indentation = max(map(len, items.keys())) + 4
    return '\n'.join(map(lambda x: '\n'.join(_indent_mapping_item(x[0], x[1], max_indentation)), items.items()))


def _indent_mapping_item(key: str, value: str, max_indentation: int) -> Iterator[str]:
    lines = value.split('\n')
    yield f'{key}{" " * (max_indentation - len(key))}    {lines[0]}'
    for line in lines[1:]:
        yield f'{" " * max_indentation}    {line}'


def report() -> str:
    return _indent_mapping({
        'Betty': version_label(),
        'Operating system': platform.platform(),
        'Python': sys.version,
        'Python packages': _indent_mapping({
            x.metadata['Name']: x.version
            for x
            in sorted(
                distributions(),
                key=lambda x: x.metadata['Name'].lower(),  # type: ignore[no-any-return, unused-ignore]
            )
        }),
    })



from __future__ import annotations

import asyncio
from asyncio import TaskGroup
from functools import wraps
from threading import Thread
from typing import Callable, Awaitable, TypeVar, Generic, cast, ParamSpec, Coroutine, Any

P = ParamSpec('P')
T = TypeVar('T')


async def gather(*coroutines: Coroutine[Any, None, T]) -> tuple[T, ...]:
    tasks = []
    async with TaskGroup() as task_group:
        for coroutine in coroutines:
            tasks.append(task_group.create_task(coroutine))
    return tuple(
        task.result()
        for task
        in tasks
    )


def wait(f: Awaitable[T]) -> T:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None
    if loop:
        synced = _SyncedAwaitable(f)
        synced.start()
        synced.join()
        return synced.return_value
    else:
        return asyncio.run(
            f,  # type: ignore[arg-type]
        )


def sync(f: Callable[P, Awaitable[T]]) -> Callable[P, T]:
    @wraps(f)
    def _synced(*args: P.args, **kwargs: P.kwargs) -> T:
        return wait(f(*args, **kwargs))
    return _synced


class _SyncedAwaitable(Thread, Generic[T]):
    def __init__(self, awaitable: Awaitable[T]):
        super().__init__()
        self._awaitable = awaitable
        self._return_value: T | None = None
        self._e: BaseException | None = None

    @property
    def return_value(self) -> T:
        if self._e:
            raise self._e
        return cast(T, self._return_value)

    @sync
    async def run(self) -> None:
        try:
            self._return_value = await self._awaitable
        except BaseException as e:
            self._e = e



from __future__ import annotations

import hashlib
import json
import logging
import re
from contextlib import suppress
from os.path import getmtime
from pathlib import Path
from time import time
from typing import cast, Any

import aiofiles
import aiohttp

from betty.app import App
from betty.asyncio import gather
from betty.functools import filter_suppress
from betty.locale import Localized, negotiate_locale, to_locale, get_data, LocaleNotFoundError, Localey
from betty.media_type import MediaType
from betty.model import Entity
from betty.model.ancestry import Link, HasLinks


class WikipediaError(BaseException):
    pass


class NotAnEntryError(WikipediaError, ValueError):
    pass


class RetrievalError(WikipediaError, RuntimeError):
    pass


_URL_PATTERN = re.compile(r'^https?://([a-z]+)\.wikipedia\.org/wiki/([^/?#]+).*$')


def _parse_url(url: str) -> tuple[str, str]:
    match = _URL_PATTERN.fullmatch(url)
    if match is None:
        raise NotAnEntryError
    return cast(tuple[str, str], match.groups())


class Entry(Localized):
    def __init__(self, locale: str, name: str, title: str, content: str):
        super().__init__(locale=locale)
        self._name = name
        self._title = title
        self._content = content

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return 'https://%s.wikipedia.org/wiki/%s' % (self.locale, self._name)

    @property
    def title(self) -> str:
        return self._title

    @property
    def content(self) -> str:
        return self._content


class _Retriever:
    def __init__(self, http_client: aiohttp.ClientSession, cache_directory_path: Path, ttl: int = 86400):
        self._cache_directory_path = cache_directory_path
        self._cache_directory_path.mkdir(exist_ok=True, parents=True)
        self._ttl = ttl
        self._http_client = http_client

    async def _request(self, url: str) -> Any:
        cache_file_path = self._cache_directory_path / hashlib.md5(url.encode('utf-8')).hexdigest()

        response_data = None
        with suppress(FileNotFoundError):
            if getmtime(cache_file_path) + self._ttl > time():
                async with aiofiles.open(cache_file_path, encoding='utf-8') as f:
                    json_data = await f.read()
                response_data = json.loads(json_data)

        if response_data is None:
            logger = logging.getLogger()
            try:
                async with self._http_client.get(url) as response:
                    response_data = await response.json(encoding='utf-8')
                    json_data = await response.text()
                    async with aiofiles.open(cache_file_path, 'w', encoding='utf-8') as f:
                        await f.write(json_data)
            except aiohttp.ClientError as e:
                logger.warning('Could not successfully connect to Wikipedia at %s: %s' % (url, e))
            except ValueError as e:
                logger.warning('Could not parse JSON content from Wikipedia at %s: %s' % (url, e))

        if response_data is None:
            try:
                async with aiofiles.open(cache_file_path, encoding='utf-8') as f:
                    json_data = await f.read()
                response_data = json.loads(json_data)
            except FileNotFoundError:
                raise RetrievalError('Could neither fetch %s, nor find an old version in the cache.' % url)

        return response_data

    async def _get_page_data(self, url: str) -> Any:
        response_data = await self._request(url)
        try:
            return response_data['query']['pages'][0]
        except (LookupError, TypeError) as e:
            raise RetrievalError('Could not successfully parse the JSON format returned by %s: %s' % (url, e))

    async def get_translations(self, entry_language: str, entry_name: str) -> dict[str, str]:
        url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (
            entry_language, entry_name)
        page_data = await self._get_page_data(url)
        try:
            translations_data = page_data['langlinks']
        except KeyError:
            # There may not be any translations.
            return {}
        return {translation_data['lang']: translation_data['title'] for translation_data in translations_data}

    async def get_entry(self, language: str, name: str) -> Entry:
        url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=extracts&exintro&format=json&formatversion=2' % (
            language, name)
        page_data = await self._get_page_data(url)
        try:
            return Entry(language, name, page_data['title'], page_data['extract'])
        except KeyError as e:
            raise RetrievalError('Could not successfully parse the JSON content returned by %s: %s' % (url, e))


class _Populator:
    def __init__(self, app: App, retriever: _Retriever):
        self._app = app
        self._retriever = retriever

    async def populate(self) -> None:
        locales = set(map(lambda x: x.alias, self._app.project.configuration.locales.values()))
        await gather(*(
            self._populate_entity(entity, locales)
            for entity
            in self._app.project.ancestry
        ))

    async def _populate_entity(self, entity: Entity, locales: set[str]) -> None:
        if not isinstance(entity, HasLinks):
            return

        entry_links: set[tuple[str, str]] = set()
        for link in entity.links:
            try:
                entry_locale, entry_name = _parse_url(link.url)
            except NotAnEntryError:
                continue
            else:
                try:
                    get_data(entry_locale)
                except LocaleNotFoundError:
                    continue
                else:
                    entry_links.add((entry_locale, entry_name))

            entry = None
            if link.label is None:
                with suppress(RetrievalError):
                    entry = await self._retriever.get_entry(entry_locale, entry_name)
            await self.populate_link(link, entry_locale, entry)

        for entry_locale, entry_name in list(entry_links):
            entry_translations = await self._retriever.get_translations(entry_locale, entry_name)
            if len(entry_translations) == 0:
                continue
            entry_translation_locale_datas: set[Localey] = set(filter_suppress(get_data, LocaleNotFoundError, entry_translations.keys()))
            for locale in locales.difference({entry_locale}):
                added_entry_locale_data = negotiate_locale(locale, entry_translation_locale_datas)
                if added_entry_locale_data is None:
                    continue
                added_entry_locale = to_locale(added_entry_locale_data)
                added_entry_name = entry_translations[added_entry_locale]
                if (added_entry_locale, added_entry_name) in entry_links:
                    continue
                try:
                    added_entry = await self._retriever.get_entry(added_entry_locale, added_entry_name)
                except RetrievalError:
                    continue
                added_link = Link(added_entry.url)
                await self.populate_link(added_link, added_entry_locale, added_entry)
                entity.links.add(added_link)
                entry_links.add((added_entry_locale, added_entry_name))

    async def populate_link(self, link: Link, entry_locale: str, entry: Entry | None = None) -> None:
        if link.url.startswith('http:'):
            link.url = 'https:' + link.url[5:]
        if link.media_type is None:
            link.media_type = MediaType('text/html')
        if link.relationship is None:
            link.relationship = 'external'
        if link.locale is None:
            link.locale = entry_locale
        if link.description is None:
            # There are valid reasons for links in locales that aren't supported.
            with suppress(ValueError):
                link.description = self._app.localizers.get_negotiated(link.locale)._('Read more on Wikipedia.')
        if entry is not None and link.label is None:
            link.label = entry.title
